from typing import TypeVar

from rainbowadn.data.atomic.integer import Integer
from rainbowadn.data.collection.trees.binary.binarytreebalancingprotocol import BinaryTreeBalancingProtocol
from rainbowadn.data.collection.trees.binary.binarytreecreationprotocol import BinaryTreeCreationProtocol
from rainbowadn.core.hashpoint import HashPoint

__all__ = ('AVLBTBP',)

ActiveKeyType = TypeVar('ActiveKeyType')
TreeType = TypeVar('TreeType')


class AVLBTBP(BinaryTreeBalancingProtocol[ActiveKeyType, Integer, TreeType]):
    def empty_metadata(self) -> HashPoint[Integer]:
        return HashPoint.of(Integer(0))

    def metadata(
            self,
            treel: TreeType,
            treer: TreeType,
            key: HashPoint[ActiveKeyType],
            protocol: BinaryTreeCreationProtocol[ActiveKeyType, Integer, TreeType]
    ) -> HashPoint[Integer]:
        return HashPoint.of(
            Integer(1 + max(self.height(treel, protocol), self.height(treer, protocol)))
        )

    @classmethod
    def height(
            cls,
            tree: TreeType,
            protocol: BinaryTreeCreationProtocol[ActiveKeyType, Integer, TreeType]
    ) -> int:
        if (split := protocol.split(tree)) is None:
            return 0
        else:
            return split.metadata.resolve().integer

    def balance(
            self,
            tree: TreeType,
            protocol: BinaryTreeCreationProtocol[ActiveKeyType, Integer, TreeType]
    ) -> TreeType:
        if (split := protocol.split(tree)) is None:
            return tree
        else:
            delta = self.height(split.treel, protocol) - self.height(split.treer, protocol)
            assert isinstance(delta, int)
            if delta < -1:
                assert (splitr := protocol.split(split.treer)) is not None
                if self.height(splitr.treel, protocol) > self.height(splitr.treer, protocol):
                    assert (splitrl := protocol.split(splitr.treel)) is not None
                    return protocol.tree(
                        protocol.tree(split.treel, splitrl.treel, split.key),
                        protocol.tree(splitrl.treer, splitr.treer, splitr.key),
                        splitrl.key
                    )
                else:
                    return protocol.tree(
                        protocol.tree(split.treel, splitr.treel, split.key),
                        splitr.treer,
                        splitr.key
                    )
            elif delta > 1:
                assert (splitl := protocol.split(split.treel)) is not None
                if self.height(splitl.treer, protocol) > self.height(splitl.treel, protocol):
                    assert (splitlr := protocol.split(splitl.treer)) is not None
                    return protocol.tree(
                        protocol.tree(splitl.treel, splitlr.treel, splitl.key),
                        protocol.tree(splitlr.treer, split.treer, split.key),
                        splitlr.key
                    )
                else:
                    return protocol.tree(
                        splitl.treel,
                        protocol.tree(splitl.treer, split.treer, split.key),
                        splitl.key
                    )
            else:
                return tree
